﻿#define sharedContext

using System;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


namespace Threading
{
    // async methods cannot have out or ref parameters
    
    public partial class MainForm : Form
    {
#if sharedContext
        SynchronizationContext m_syncContext; // http://www.codeproject.com/Articles/31971/Understanding-SynchronizationContext-Part-I
#endif
        SendOrPostCallback m_UIUpdateMethod;

        private SynchronizationContext Context
        {
            get
            {
                return
#if sharedContext
                    m_syncContext;
#else
                    SynchronizationContext.Current;
#endif
            }
        }

        // This method updates UI - runs on Main thread
        private void UIUpdateMethod(object state)
        {
           listBox1.Items.Add((string)state);
           // Thread.Sleep(3000); this will block UI when called with either Send or Post
        }

        private void Log(string data)
        {
            m_syncContext.Post(UIUpdateMethod, GetMessageHeader() + "  " + data);
            // Context.Send() will hang when for Task demo example
        }

        public MainForm()
        {
            InitializeComponent();
#if sharedContext
            m_syncContext = SynchronizationContext.Current;
#endif
            m_UIUpdateMethod = UIUpdateMethod;

            label1.Text = "UI Thread id: "+ Thread.CurrentThread.ManagedThreadId.ToString();
        }

        private string GetMessageHeader()
        {
            return DateTime.Now.ToLongTimeString() + " Thread id:" + Thread.CurrentThread.ManagedThreadId;
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            listBox1.Items.Clear();
        }

        //-----------------------------------------------------------------------------------------
        // Task   Task.Start(), Task.Wait(), Task.RunSynchronously()
        //-----------------------------------------------------------------------------------------

        private void LongTaskNoParam()
        {
            int sleepTime = 1000;
            Log(string.Format(" start Task={0}, obj={1}", Task.CurrentId, sleepTime.ToString()));
            Thread.Sleep((int)sleepTime);
            Log(string.Format(" end   Task={0}, obj={1}", Task.CurrentId, sleepTime.ToString()));
        }

        private void LongTaskParam(object sleepTime)
        {
            Log(string.Format(" start Task={0}, obj={1}", Task.CurrentId, sleepTime.ToString()));
            Thread.Sleep((int)sleepTime);
            Log(string.Format(" end   Task={0}, obj={1}", Task.CurrentId, sleepTime.ToString()));
        }

        private void btTask_click(object sender, EventArgs e)
        {
            Log("UI will be blocked during the execution");

            // Create and start new task on a worker thread
            Task task1 = new Task(LongTaskNoParam);
            task1.Start();
            Application.DoEvents();

            // Create and start new task on a worker thread with fatcory
            Task task2 = Task.Factory.StartNew(LongTaskNoParam);

            // Pass argument to the worker method 
            Task task3 = Task.Factory.StartNew(LongTaskParam, 3000);

            // Wait until the task to finishes 
            // UI willbe blocked if you wait in the same method
            // you can limit execution by timeout or pass cancellation token
            task3.Wait();
            Log("task3 has finished");
            Application.DoEvents();

            // task3.RunSynchronously(); will fail to start second time
            // start the task on the same thread
            Task task4 = new Task(LongTaskParam, 2000);
            task4.RunSynchronously();
            Log("task4 has finished synchronously");
        }

        //-----------------------------------------------------------------------------------------
        // async - await Basics
        //-----------------------------------------------------------------------------------------

        public async Task<string> RunsAsynchronously(int sleepTime)
        {
            Log(string.Format("before: await Task.Delay({0});", sleepTime));
            await Task.Delay(sleepTime);
            Log(string.Format(" after: await Task.Delay({0});", sleepTime));
            return " Finished Async";
        }

        public async Task<string> RunsAsynchronouslyWithoutWait(int sleepTime)
        {
            Log(string.Format("before:X await Task.Delay({0});", sleepTime));
            Task.Delay(sleepTime);
            Log(string.Format(" after:X await Task.Delay({0});", sleepTime));
            return " Finished Async";
        }

        public async Task<string> RunsWithUIBlocked()
        {
            Log("before: Thread.Sleep(2000);");
            Thread.Sleep(2000); // this block UI
            Log(" after: Thread.Sleep(2000);");
            return "Finished sync";
        }

        private async void btAsyncAwaitBasics_Click(object sender, EventArgs e)
        {
            // all the code here and in called methods runs on the same thread
            string s;
            
            // waits here and in inside RunsAsynchronously
            Log("before: RunsAsynchronously(3000);");
            s = await RunsAsynchronously(3000); 
            Log(" after: RunsAsynchronously(3000);");

            // does NOT wait in this method, waits in RunsAsynchronously
            Log("before: RunsAsynchronously(5000);");
            RunsAsynchronously(5000); 
            Log(" after: RunsAsynchronously(5000);");

            // does NOT wait inside RunsAsynchronously, waits here, but for fraction of second
            Log("before: RunsAsynchronouslyWithoutWait(4000);");
            s = await RunsAsynchronouslyWithoutWait(4000);  
            Log(" after: RunsAsynchronouslyWithoutWait(4000);");

            // waits, but UI will not be responsive
            Log("before: RunsSynchronously(); - UI will not be responsive");
            Application.DoEvents();
            s = await RunsWithUIBlocked(); 
            Log(" after: RunsSynchronously();");
        }

        //-----------------------------------------------------------------------------------------
        // Calls returning void and type from async and non-async methods
        //-----------------------------------------------------------------------------------------

        // the methods below demonstrates legacy code WITHOUT async signature 
        private void LongTask(int sleepTime)
        {
            Log(string.Format(" LongTask() {0} start", sleepTime));
            Thread.Sleep(sleepTime);
            Log(string.Format(" LongTask() {0} end", sleepTime));
        }

        private string LongTaskWithResult(int sleepTime)
        {
            Log(string.Format(" LongTaskWithResult() {0} start", sleepTime));
            Thread.Sleep(sleepTime);
            Log(string.Format(" LongTaskWithResult() {0} end", sleepTime));
            return "LongTaskWithResultVALUE";
        }


        // the methods below represents wrapper that allows calling legacy code asynchronously
        private async Task LongTaskAsync(int sleepTime)
        {
            await Task.Factory.StartNew(() => LongTask(sleepTime));
        }

        private async Task<string> LongTaskWithResultAsync(int sleepTime)
        {
            string s = await Task.Factory.StartNew(() => LongTaskWithResult(sleepTime));
            return s;
        }


        private void CallLegacyCodeFromNonAsyncMethodTest()
        {
            // all the calls below start tasks on a "worker" thread
            
            // these calls do NOT wait for LongTask() to complete
            Task.Factory.StartNew( () => { LongTask(2000); } );
            Log(" after () => { LongTask(2000); } ");

            Task.Factory.StartNew( () => { LongTaskWithResult(3000); });
            Log(" after () => { LongTaskWithResult(3000); } ");

            
            // these calls wait for LongTask() to complete
            Task.Factory.StartNew(async () =>
            {
                await LongTaskAsync(4000);  // needs to call the wrapper method
                Log(" after () => { await LongTask(4000); } ");
            });

            string s;
            Task.Factory.StartNew(async () =>
            {
                s = await LongTaskWithResultAsync(5000);  // needs to call the wrapper method
                Log(" after () => { await LongTaskWithResult(5000); } ");
            });
        }

        private void btCallLegacyCodeFromNoAsync_Click(object sender, EventArgs e)
        {
            //-- calling from NON ASYNC method
            CallLegacyCodeFromNonAsyncMethodTest();

        }

        private async void CallLegacyCodeFromAsyncMethodTest()
        {
            //// A. method returning void

            // calling LongTask() via wrapper
            await LongTaskAsync(2000);
            Log("after: await LongTaskAsync(2000)");

            // calling LongTask() directly
            await Task.Factory.StartNew( /* async */ () =>    // async can be used here but can be omitted (if the this method is marked with async)
            {
                // textBox1.Text += "bla";    it is not allowed to update UI from worker thread. error: cross thread operation not valid
                LongTask(3000); 
            });
            Log("after: await await Task.Factory.StartNew(() => LongTask(3000));");


            //// B. method returning result

            // calling LongTaskWithResult() via wrapper
            string s = await LongTaskWithResultAsync(2500);
            Log("after: await LongTaskWithResultAsync(2500);");

            // calling LongTaskWithResult() directly
            s = await Task<string>.Factory.StartNew(/* async */ () =>  // async can be used here but can be omitted (if the this method is marked with async)
            { 
                return LongTaskWithResult(3500); 
            });
            Log("after: await Task<string>.Factory.StartNew(() => { return LongTaskWithResult(3500); });");
        }


        private void btCallLegacyCodeFromAsync_click(object sender, EventArgs e)
        {
            //-- calling from ASYNC method
            CallLegacyCodeFromAsyncMethodTest();
        }

       
        //-----------------------------------------------------------------------------------------
        // Calling chain - Nested await keywords
        //-----------------------------------------------------------------------------------------
        private async void btCallingChain_Click(object sender, EventArgs e)
        {
            Task.Factory.StartNew( async () => 
            {
                Log("before call");
                await LongTaskAsync(2000);
                Log("after call");    
            });
        }


        //-----------------------------------------------------------------------------------------
        // Cancel async call
        //-----------------------------------------------------------------------------------------
        private void LongTaskCancelCapable(CancellationToken ct)
        {
            Log("start");
            for (int i = 0; i < 100; i++)
            {
                Log(" cycle " + Convert.ToString(i));
                if (ct.IsCancellationRequested)
                {
                    // cancel execution
                    Log("Cancelled");
                    return;
                }
                else
                {
                    // normal processing
                    Thread.Sleep(1000);  // UI is not blocked even if Thread.Sleep() is used
                }
            }
        }

        private async void btCancelExec_Click(object sender, EventArgs e)
        {
            // async is used to allow Task.Delay() otherwise UI will not be refreshed
            
            CancellationTokenSource cts = new CancellationTokenSource();
            CancellationToken ct = cts.Token;

            // start long running task that is capable to be cancelled
            Task.Factory.StartNew( /* async */ () => LongTaskCancelCapable(ct), ct);
            
            // wait some time to see that long running task is running
            await Task.Delay(3000);
            // Thread.Sleep(3000);  // uncomment to see what will happen when Thread.Sleep() is called; you can remove "async" as well
            
            // cancel the long running task
            cts.Cancel();
            Log("Cancellation requested");
        }





        /*
         *  no async Thread.Sleep  - UI blocked - classic
         *  async Thread.Sleep - UI still blocked 
         *  Task.Delay  - does not wait
         *  await Task.Delay - waits and dpes not cblock UI
         * 
         * 
         * 
         * Method runs synchronusly until await first is reached.
         * 
         * 
         * Task.Delay() starts waiting loop on another thread
         * 
         */


    }
}
